#this might not need it's own file lmao

import csv

#i don't think this has been called
def read(fname):
    return read_str(open(fname, "r").read())

#no need for a lexer as the csvreader does it automatically
def read_str(string):
    tmp = [n for n in csv.reader(string.split("\n"), delimiter=" ")]
    return tmp



from macro_lib import macro

#this file tests the capabilities for the macro library to pass arguments

def trace(tups):
    for n in tups: print(n)
    return tups

@macro(trace)
def test(a):
    print("a passed as " + str(a))
    return a
print(test(3))



from macro_lib import macro

def constant(tups_in):
    ret = []
    constants = {}
    for n in tups_in:
        inst, arg = n
        if inst == "LOAD_CONST" and len(arg) > 6 and arg[1:6] == "const":
            var_name, val = arg[7:-1].split(" ")
            if var_name not in constants:
                constants[var_name] = val
            else:
                raise Exception("tried to redefine a constant")
        elif arg in constants:
            ret.append(("LOAD_CONST", constants[var_name]))
        else:
            ret.append(n)
    return ret

@macro(constant)
def test():
    _ = "const x 15" #note: x is only available in this scope
    print(x)
    x = 100
    print(x) #15 because x is constant set to 15
    #_ = "const x 100" #error because you cannot redefine a constant

test()



#this is some of the globals that are added to the code object in compilation
#many of them need to get affected by some of the instructions so I have made them mutable
#globals
    #THIS IS UGLY CODE AND IF I COULD THINK OF A CLEANER WAY TO DO THIS I WOULD BE THIS IS THE
    #ONLY WAY I CAN REALLY THINK OF THAT IS CLEAR AND OBVIOUS
        #woops caps lock
        #i could do this with my own code object and messing wit hthat but that would maybe be better code?
        #NOTE: potentially change this to be using a struct to store data rather than globals

#this file is only as long as it is cause of my ranting comments

from functools import reduce, partial
from bytecodemacro.bytecodecompiler.CodeContext import CodeContext

#the public preprocessing to be called
    #i'm not sure if chucking everything into a function like this is good or bad practice but i'm
    #doing it to keep my algorithm nice and clean
#this function could also be conceptualized as compiling an intermediary language (which is closer
            #to the output of dis)
def PreProcess(code_context_in, all_objects):

    #the naming schema is a bit weird with these lists but it is to keep it consistent with the
    #documentations (in the dis library) names
    constants = [] #list of literals used in the program (i would use a set for faster seraching but the instructions rely on the index of the literal so i have to 
    names = [] #list of names used in the program (same logic about sets as above)
    varnames = code_context_in.varnames #( the arguments already exist as varnames)
    global_list = [] #not called globals cause its reserved (also not sure it's necersarry, i'm a bit confused about store_global
    cellvars = [] #not sure what this is

    #below this line is globals that aren't returned but just used in preprocessing
    compare_list = ('<', '<=', '==', '!=', '>', '>=') 
    label_positions = {}

    #used to compose the pre-processing functions together
    def compose_list(f_list):
        def ret(arg):
            #probably a better way to do this (like a fold probaly?)
            return reduce(lambda a,n: n(a), f_list, arg)
        return ret

    def bind(f, li):
        return list(reduce(lambda a1,a2: a1+a2, map(f, li))) #not fast but it's elegant

    def should_remove(line): #called in the filter function
        if len(line) == 0: return False #this is already done in parse but i've kept it here just in
                                        #case i delete it
        if line[0][0] == "#": return False
        return True

    def find_labels(lines):
        #not a massive fan of this function but it works
            #i could do something with 
        ret = []
        for i, v in enumerate(lines):
            if len(v) == 2 and v[0] == "LABEL":
                label_positions[v[1]] = i*2 #*2 because it is the position in the byte string and
                                              #it is 2 bytes per instruction
                v = ["NOP"]
            ret+=[v]
        return ret

    #removes any empty arguments
    def clean_args(val):
        if len(val) == 1: return val
        if val[1] == '':  return [val[0]]
        else:             return val[:2]

    #for instrucitons that don't have arguments (like pop_top or return_value)
    def add_arg(val):
        if len(val) == 1: 
            return val + [0]
        return val

    def memoize_loading(lis, arg): #note: memoize loading is called sometimes with things that should not be added to 
        try: 
            arg = int(arg)
        except ValueError:
            try:
                arg = float(arg)
            except ValueError: pass
        finally:
            if arg in lis: arg = lis.index(arg) #double check later
            else: lis.append(arg); arg = len(lis)-1 #maybe remove semi colon
            return (lis, arg)

    #commands to lists is used in handle_literals
        #note: i am not sure how load_closurer works and it has something weird about freevars in it
    commands_to_lists = {"LOAD_CONST": constants, "LOAD_NAME": names, "LOAD_GLOBAL":
            global_list, "LOAD_FAST": varnames, "STORE_FAST": varnames, "DELETE_FAST":
            varnames, "LOAD_ATTR": names,
            "LOAD_CLOSURE": cellvars, "LOAD_DEREF": cellvars, "LOAD_CLASSDEREF": cellvars,
            "DELETE_DEREF": cellvars, "LOAD_METHOD": names, "STORE_NAME": names,
            "STORE_ATTR": names, "STORE_GLOBAL": global_list, "STORE_DEREF": cellvars,
            "DELETE_NAME": names, "DELETE_ATTR": names, "DELETE_GLOBAL": global_list,
            "IMPORT_NAME": names, "IMPORT_FROM": names, "COMPARE_OP": compare_list}
    def handle_literals(val):
        inst, arg = val
        if inst in commands_to_lists: #might have to nonlocal the commands_to_lists
            lis = commands_to_lists[inst]
            lis, arg = memoize_loading(lis, arg)
        return (inst, arg)

    #references a dictionary not a list so it has to be done like this
    #i have to handle jump_relatives (e.g: jump_forward) in a different function
    def handle_jump_absolute_literals(line):
        inst, arg = line
        if inst in {"POP_JUMP_IF_TRUE", "POP_JUMP_IF_FALSE", "JUMP_IF_NOT_EXC_MATCH", "JUMP_IF_TRUE_OR_POP", "JUMP_IF_FALSE_OR_POP", "JUMP_ABSOLUTE"}:
            arg = label_positions[arg]
        return (inst, arg)

    obj_dict = {o[0].name: o for o in all_objects}
    def handle_load_object(line):
        nonlocal constants
        inst, arg = line
        if inst == "LOAD_OBJECT":
            inst = "LOAD_CONST"
            constants, arg = memoize_loading(constants, obj_dict[arg])
            inst2 = "LOAD_CONST"
            constants, arg2 = memoize_loading(constants, 0) #i should be calling handle_literals instead of this, but tbh at this point i just don't care
            #the reason we have objects stored in a list and then subscripted is too emulate pointer
            #behaviour for later on so we can have cyclic references and the like, if you want a
            #more detailed explanation, give us an email: jonte.fry@gmail.com
            return [(inst, arg), (inst2, arg2), ("BINARY_SUBSCR",0)]
        else:
            return [(inst, arg)]

    def handle_jump_relative_literals(lines):
        ret = []
        for i, v in enumerate(lines):
            inst, arg = v
            if inst in {"SETUP_WITH", "JUMP_FORWARD", "FOR_ITER", "SETUP_FNIALLY"}:
                arg = label_positions[arg] - i*2 #i*2 as there are two bytes per instruction
            ret += [(inst, arg)]
        return ret

    #used in debugging (delete it later)
    def trace(v): #can't put in lambda cause one line
        print(v)
        return v

    def increment_labels_below(index, n):
        #sometimes insturctions need to be added after labels defined so they need to be updates
        nonlocal label_positions
        label_positions = {key: val if val < index else val + n*2 for key, val in label_positions.items()} #n * 2 as instructions are 2 bytes long

    #terrible name for a function
        #it's too early too think of anything good though
    def extended_args_instructions(n):
        if n < 256: return [("EXTENDED_ARG", n)]
        if n > 255: 
            tmp = ("EXTENDED_ARG", n % 256)
            return [tmp] + extended_args_instructions(n >> 8)

    #sometimes the arguments to an instruction become bigger then 255 (a byte) (particularly in jump
    #commands) so i have to add a thing to auto generate the extended args
        #i'm not a massive fan of this function, too much maths and i'm relying on calling the
        #funciton to execute some of the logic rather than the function wholeheartedly
        #i need to change this later
    def handle_extended_args(lines):
        ret = []
        for i, v in enumerate(lines):
            inst, arg = v 
            if arg > 255: 
                e = extended_args_instructions(arg >> 8)
                increment_labels_below(i, len(e))
                arg = arg % 256
                ret += e
            ret += [(inst, arg)]
        return ret

    #maybe not the cleanest to do it like this but it's more fun like this
        #the order of the functions here is not completely arbitrary but i can't be botehred figuring out the best way to do it
    pre_process = compose_list([
        partial(filter, should_remove),
        find_labels, #this can't be in the map cause it needs to have index information
        partial(map, compose_list([
            clean_args,
            add_arg, 
            lambda n: (n[0].upper(), n[1]), #makes the instructions uppercase (has to be a lambda cause .upper is a method (ruins my pointfree >:(
            handle_literals,
            handle_jump_absolute_literals
        ])), #composes all the preproccessing that only needs a single line and not the context of other lines
        partial(bind, handle_load_object),
        handle_jump_relative_literals,
        partial(map, lambda n: (n[0], int(n[1]))),
        handle_extended_args,
        list #list might not be necersarry, but it's better safe than sorry in future updates
    ])

    #look what you have done to me oo
    code_context_in.bytes = pre_process(code_context_in.bytes)
    code_context_in.constants = constants
    code_context_in.names = names
    code_context_in.global_list = global_list
    code_context_in.cellvars = cellvars
    code_context_in.varnames = varnames



import marshal
from types import CodeType
from bytecodemacro.bytecodecompiler.Compile import Compile #in Compile.py (not a library)

class CodeContext:
    """
        this class holds the data for a piece of intermediary bytecode along with the context surrounding it
    """
    def __init__(self, **kwargs):
        #need to handle non existent key in kwargs
        self.arg_num = 0 if "arg_num" not in kwargs else kwargs["arg_num"] #might not be necersarry
        #skipped kwargnum (in original code object specification
        #skiped something unknown
        #skipped number of locals
        #skipped stacksize
        #skipped flag
        self.bytes = [] if "bytestring" not in kwargs else kwargs["bytestring"]
        self.constants = [] if "constants" not in kwargs else kwargs["constants"]
        self.names = [] if "names" not in kwargs else kwargs["names"]
        #skipped something unknown
        #skipped fname
        self.fname = "" if "fname" not in kwargs else kwargs["fname"] #might not be necersarry, still a lil foggy on the details
        self.name = "" if "name" not in kwargs else kwargs["name"]
        #skipped first line unmber
        #skipped real code line count array
        #skipped something unkown
        #skipped something unknown
        #below this line are things that are probably in the original code object but i'm not
        #sure where
        self.varnames = [] if "varnames" not in kwargs else kwargs["varnames"]
        self.globals = [] if "global_list" not in kwargs else kwargs["global_list"]
        self.cellvars = [] if "cellvars" not in kwargs else kwargs["cellvars"]
        #self.args = [] if "args" not in kwargs else kwargs["args"]

    def MakePyObj(self): #should be called after object is generated from PreProcess.py
        codestring = Compile(self.bytes)
        return CodeType(
                self.arg_num, #arg count
                0, #kwarg count
                0, #still have no clue what this is or how i figured this out
                len(self.varnames), #might not be right
                50000, #note: can't figure this out cause halting problem
                0, #flags for the compiler, (i will need to take a double check at this later)
                codestring,
                tuple(self.constants),
                tuple(self.names),
                tuple(self.varnames),
                "Generated_by_GlubsCPyAsm",
                self.name,
                1,
                bytes([1]*len(self.bytes)), #could be generated a list of ones but tbh don't really matter
                (), #gonna handle freevars later (maybe idk i'm confused (might not be necersarry to hand write (maybe for later compilers))
                tuple(self.cellvars)
        )



#this is an example use case of the macros library to implement gotos in python
#only reason this is not in it's own folder is to make importing easier
from macro_lib import macro

def goto(tups):
    ret = []
    for n in tups:
        inst, arg = n
        if inst == "LOAD_CONST" and len(arg) > 6 and arg[1:5] == "goto":
            ret.append(("JUMP_ABSOLUTE", arg[6:-1]))
        elif inst == "LOAD_CONST" and len(arg) > 2 and arg[-2] == ":":
            ret.append(("LABEL", arg[1:-2]))
        else:
            ret.append(n)
    return ret

@macro(goto)
def f():
    n = 0
    while n < 10:
        if n > 5:
            _ = "goto exit"
        print(n)
        n+=1
    _ = "exit:"
f()



from macro_lib import macro, byte_compile, unused_var

#i really gotta do something about all these if statements
    #also this only works with strings but fuck it, thats cool as right now
def walrus(tups):
    ret = []
    for n in tups:
        inst, arg = n
        if inst == "LOAD_CONST":
            #poorly named variables
            tmp = arg.split(" ")
            if len(tmp) >= 2 and tmp[1] == ":=":
                tmp2 = byte_compile(" ".join(tmp[2:-1]) + " " + tmp[-1][:-1]) #last one removes the extra " (might create extra space in string?)
                if len(tmp2[-1]) > 1 and tmp2[-1][1] == "__doc__": tmp2 = tmp2[:-1] #idk whats happening here
                ret += tmp2
                var_name = unused_var(tups, ret)
                ret.append(("STORE_FAST", var_name)) #change to new_var later
                ret.append(("LOAD_FAST", var_name)) #change to new_var later
                ret.append(("STORE_NAME", tmp[0][1:])) #remove extra "
                ret.append(("LOAD_FAST", var_name)) #adds extra walrus to stack
            else:
                ret.append(n)
        else:
            ret.append(n)
    print(ret)
    return ret

@macro(walrus)
def test(): #should print hello world twice
    print("x := 'hello world'")
    if x == "hello world":
        print("walrus successful!")
    else:
        print("oops")
test()



import types

#update to add inner function and class support (idk about classes i just haven't tested it)
def get_objects(obj): 
    ret = [n for n in obj.co_consts if isinstance(n, types.CodeType)]
    return [obj] + ret

def var_names(obj):
    return [obj.co_varnames[n] for n in range(obj.co_argcount)]

def add_definitions(obj, tuples): #tuples have already been pre processed
    out = [("DEFINE", "main" if obj.co_name == "<module>" else obj.co_name)]
    out += [("ADD_ARG", n) for n in var_names(obj)]
    out += tuples
    out += [("END", "0")]
    return out



#this is an attempt to build a c style for loop in python

from macro_lib import macro, byte_compile

def split(string_in):
    return string_in.split(";")

def cloop(tups):
    ret = []
    for_iterator = []
    checker = []
    for n in tups:
        if len(n) > 2: #fix later
            ret.append(n)
            continue
        inst, arg = n
        if arg == "_": continue #just cause my code is bad lmao
        elif inst == "LOAD_CONST" and len(arg) > 3 and arg[1:4] == "for":
            definition, check, iterator = tuple(split(arg[5:-1]))
            ret += byte_compile(definition)
            ret.append(("LABEL", "for"))
            for_iterator = byte_compile(iterator)
            checker = byte_compile(check)[:-1] #pops the result of comparison
        elif inst == "LOAD_CONST" and arg == '"end"':
            ret += for_iterator
            ret += checker
            ret.append(("POP_JUMP_IF_TRUE", "for"))
        else:
            ret.append(n)
    return ret

@macro(cloop)
def test(a):
    _ = "for c = 0;c < 10;c+=1"
    print(c)
    _ = "end"
test(10)



from macro_lib import macro
#to test the argument handling of the macro library more rigorously

def add_arg(tups):
    ret = tups[:1]
    ret.append(("ADD_ARG", "a")) #adds extra argument a
    for n in tups[1:]: #could do with a list comprehension or map but this is easier to understand
        inst, arg = n
        if inst == "LOAD_NAME" and arg == "a":
            #replacing load_name with load_fast cause arguments have to be load_fast but the interpreter parses all mentions of a as load_name
            ret.append(("LOAD_FAST", "a")) 
        else:
            ret.append(n)
    return ret

@macro(add_arg)
def test(b):
    print("b passed as " + str(b))
    print("a passed as " + str(a))

test(3, 4)



import dis
import marshal
import bytecodemacro.bytecodecompiler.Header as Header

def Compile(BytesStr):
    ret = []
    for n in BytesStr:
        ret += [dis.opmap[n[0]], n[1]]
    return bytes(ret)

def Out(fname, obj):
    fout = open(fname, "wb+")
    fout.write(Header.Header())
    marshal.dump(obj, fout)



from functools import reduce

def bind(f):
    def ret(li):
        return reduce(lambda n1,n2: n1+n2, map(f, li))
    return ret

def compose(inp, *fs):
    for n in fs:
        inp = n(inp)
    return inp

test_str = "<code object " #not the most rigorous check but I think it will work for now
def is_load_object(v):
    return v[0] == "LOAD_CONST" and v[2][:len(test_str)] == test_str

jump_statements = {"POP_JUMP_IF_TRUE", "POP_JUMP_IF_FALSE", "JUMP_IF_NOT_EXC_MATCH", "JUMP_IF_TRUE_OR_POP", "JUMP_IF_FALSE_OR_POP", "JUMP_ABSOLUTE", "SETUP_WITH", "JUMP_FORWARD", "FOR_ITER", "SETUP_FNIALLY"}

from bytecodemacro.uncompile.parse_dis import find_next_whitespace
#todo: handle extended_arg instruction (i think we just delete it tbh, idk i've never seen it in compiled python)
def handle_third_arg(v):
    if len(v) == 1: return [(v[0], '0')]
    elif len(v) == 2: return [v]
    inst = v[0]
    if inst in {"SETUP_WITH", "JUMP_FORWARD", "FOR_ITER", "SETUP_FNIALLY"}:
        _, arg = v[2].split() #v[2] == "to __" where __ is the absolute jump position
        return [(inst, arg)]
    elif inst in jump_statements:
        raise Exception("these should only be 2 long")
    elif is_load_object(v):
        #if this is load_const with <code object test at>
        #replace with load_object
        arg = v[2][13:] #gets the name of the object
        arg = arg[:find_next_whitespace(arg)]
        inst = "LOAD_OBJECT" #not techincally a python bytecode thing but it is in cpyasm
        return [(inst, arg)] #not handling right now for dbeugging purposes
    elif inst in {"LOAD_CONST", "LOAD_NAME", "LOAD_GLOBAL", "LOAD_FAST", "STORE_FAST", "DELETE_FAST", "LOAD_ATTR", "LOAD_CLOSURE", "LOAD_DEREF", "LOAD_CLASSDEREF", "DELETE_DEREF", "LOAD_METHOD", "STORE_NAME", "STORE_ATTR", "STORE_GLOBAL", "STORE_DEREF", "DELETE_NAME", "DELETE_ATTR", "DELETE_GLOBAL", "IMPORT_NAME", "IMPORT_FROM", "COMPARE_OP"}:
        #need to double check each one to make sure am parsing properly
        return [(inst, v[2])]

#finds places where jumps go to and adds labels to those places
    #this is elegant :)
def handle_jumps(lines):
    jump_lines = set()
    for v in lines: #could be done with a set comprehension but this is clearer
        inst, arg = v
        if inst in jump_statements: 
            jump_lines.add(int(arg))
    ret = []
    for i in range(len(lines)):
        if i*2 in jump_lines: #i*2 because the jump statements are made by counting bytes not liens
            ret.append(("LABEL", str(i*2)))
        ret.append(lines[i])
    return ret

def remove_empty(line):
    if line[0] == '' and len(line) == 1: return []
    return [line]

#strings are given with ' ' but my cpyasm only works with " "
def handle_strings(v):
    inst, arg = v
    if arg[0] == "'":
        arg = '"' + arg[1:-1] + '"'
    return (inst, arg)

def remove_globals(v):
    inst, arg = v
    if inst == "LOAD_GLOBAL":
        inst = "LOAD_NAME"
    return (inst, arg)

#for debugging
def trace(msg):
    def ret(n):
        print(msg + ": " + str(n))
        return n
    return ret

def pre_process(instructions):
    return compose(instructions, 
        bind(remove_empty),
        bind(handle_third_arg),
        lambda n: map(handle_strings, n),
        lambda n: map(remove_globals, n), #this is just cause i'm too lazy to handle globals ec dee
        list,
        handle_jumps
    )



#the header for the python bytecode, (probably doesn't need to be it's own file)

import importlib.util

def Header():
    """
    this function returns the header to be inserted at the top of the bytecode file
    """ 
    return importlib.util.MAGIC_NUMBER + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    #the header is made of two parts, first is a magic number, unqiue to each python installation,
    #and second is a timestamp for when the file was made. For clarity I have written this as all
    #zeros
    #the rest of the bytecode file is a serialized (through the marshall library) python CodeObject
    #object. The documentation for this is intentionally vague but the (currently) working code
    #should be in and explained in the compiler
